@@ -1,5 +1,5 @@
 #!/usr/bin/python3
-# Copyright (c) 2005-2012 Canonical Ltd
+# Copyright (c) 2005-2015 Canonical Ltd
 #
 # AUTHOR:
 # Michael Vogt <mvo@ubuntu.com>
@@ -29,6 +29,7 @@ import fcntl
 import fnmatch
 import gettext
 import grp
+import io
 import locale
 import logging
 import re
@@ -41,7 +42,6 @@ import sys
 
 
 from email.message import Message
-from errno import EACCES, EPERM
 from gettext import gettext as _
 from io import StringIO
 from optparse import (
@@ -62,6 +62,7 @@ import apt_pkg
 REBOOT_REQUIRED_FILE = "/var/run/reboot-required"
 MAIL_BINARY = "/usr/bin/mail"
 SENDMAIL_BINARY = "/usr/sbin/sendmail"
+USERS = "/usr/bin/users"
 # no py3 lsb_release in debian :/
 DISTRO_CODENAME = subprocess.check_output(
     ["lsb_release", "-c", "-s"], universal_newlines=True).strip()
@@ -75,6 +76,7 @@ PID_FILE = "/var/run/unattended-upgrades.pid"
 # set from the sigint signal handler
 SIGNAL_STOP_REQUEST = False
 
+
 class LoggingDateTime:
     """The date/time representation for the dpkg log file timestamps"""
     LOG_DATE_TIME_FMT = "%Y-%m-%d  %H:%M:%S"
@@ -184,7 +186,7 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
                     continue
                 try:
                     os.close(fd)
-                    #print("closed: ", fd)
+                    # print("closed: ", fd)
                 except OSError as e:
                     # there will be one fd that can not be closed
                     # as its the fd from pythons internal diropen()
@@ -211,7 +213,7 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
         try:
             adm_gid = grp.getgrnam("adm").gr_gid
             os.fchown(logfd, 0, adm_gid)
-        except KeyError:
+        except (KeyError, OSError):
             pass
         return logfd
 
@@ -243,10 +245,12 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
         os.close(logfd)
 
     def finish_update(self):
-        self._log_in_dpkg_log("Log ended: %s\n\n" % LoggingDateTime.as_string())
+        self._log_in_dpkg_log("Log ended: %s\n\n"
+                              % LoggingDateTime.as_string())
 
     def fork(self):
-        self._log_in_dpkg_log("Log started: %s\n" % LoggingDateTime.as_string())
+        self._log_in_dpkg_log("Log started: %s\n"
+                              % LoggingDateTime.as_string())
         pid = os.fork()
         if pid == 0:
             self._fixup_fds()
@@ -357,7 +361,7 @@ def match_whitelist_string(whitelist, origin):
         # strip and unquote the "," back
         (what, value) = [s.strip().replace("%2C", ",")
                          for s in token.split("=")]
-        #logging.debug("matching '%s'='%s' against '%s'" % (
+        # logging.debug("matching '%s'='%s' against '%s'" % (
         #              what, value, origin))
         # support substitution here as well
         value = substitute(value)
@@ -379,15 +383,26 @@ def match_whitelist_string(whitelist, origin):
             raise UnknownMatcherError(
                 "Unknown whitelist entry for macher '%s' (token '%s')" % (
                     what, token))
-        #logging.debug("matching '%s'='%s' against '%s'" % (
+        # logging.debug("matching '%s'='%s' against '%s'" % (
         #              what, value, origin))
     return res
 
 
-def upgrade_normal(cache, pkgs_to_upgrade, logfile_dpkg, verbose):
+def cache_commit(cache, logfile_dpkg, verbose, iprogress=None):
+    """Commit the changes from the given cache to the system"""
+    # set debconf to NON_INTERACTIVE, redirect output
+    os.putenv("DEBIAN_FRONTEND", "noninteractive")
+
+    # only act if there is anything to do (important to avoid writing
+    # empty log stanzas)
+    if len(cache.get_changes()) == 0:
+        return True, None
+
     error = None
     res = False
-    iprogress = LogInstallProgress(logfile_dpkg, verbose)
+    if iprogress is None:
+        iprogress = LogInstallProgress(logfile_dpkg, verbose)
+
     try:
         with Unlocked():
             res = cache.commit(install_progress=iprogress)
@@ -395,12 +410,17 @@ def upgrade_normal(cache, pkgs_to_upgrade, logfile_dpkg, verbose):
         error = e
         if verbose:
             logging.exception("Exception happened during upgrade.")
+    return res, error
+
+
+def upgrade_normal(cache, logfile_dpkg, verbose):
+    res, error = cache_commit(cache, logfile_dpkg, verbose)
     if res:
         logging.info(_("All upgrades installed"))
     else:
         logging.error(_("Installing the upgrades failed!"))
-        logging.error(_("error message: '%s'") % error)
-        logging.error(_("dpkg returned a error! See '%s' for details") %
+        logging.error(_("error message: '%s'"), error)
+        logging.error(_("dpkg returned a error! See '%s' for details"),
                       logfile_dpkg)
     return res
 
@@ -467,19 +487,13 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, blacklist,
         logging.debug("applying set %s" % smallest_partition)
         rewind_cache(cache, [cache[name] for name in smallest_partition])
 
-        try:
-            iprogress = LogInstallProgress(logfile_dpkg, verbose)
-            with Unlocked():
-                res = cache.commit(install_progress=iprogress)
-                if not res:
-                    raise Exception("cache.commit() returned false")
-            cache.open()
-        except Exception as e:
+        res, error = cache_commit(cache, logfile_dpkg, verbose, install_log)
+        if error:
             if verbose:
                 logging.exception("Exception happened during upgrade.")
             logging.error(_("Installing the upgrades failed!"))
-            logging.error(_("error message: '%s'") % e)
-            logging.error(_("dpkg returned a error! See '%s' for details") %
+            logging.error(_("error message: '%s'"), error)
+            logging.error(_("dpkg returned a error! See '%s' for details"),
                           logfile_dpkg)
             return False
         to_upgrade = to_upgrade - set(smallest_partition)
@@ -507,6 +521,7 @@ def is_pkgname_in_blacklist(pkgname, blacklist):
             return True
     return False
 
+
 def is_pkgname_in_whitelist(pkgname, whitelist):
     # a empty whitelist means the user does not want to use this feature
     if not whitelist:
@@ -527,6 +542,11 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist):
             logging.debug("pkg '%s' now marked delete" % pkg.name)
             return False
         if pkg.marked_install or pkg.marked_upgrade:
+            # apt will never fallback from a trusted to a untrusted
+            # origin so its good enough if we have a single trusted one
+            if not any([o.trusted for o in pkg.candidate.origins]):
+                logging.debug("pkg '%s' is untrusted" % pkg.name)
+                return False
             if not is_allowed_origin(pkg.candidate, allowed_origins):
                 logging.debug("pkg '%s' not in allowed origin" % pkg.name)
                 return False
@@ -534,14 +554,15 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist):
                 logging.debug("pkg '%s' package has been blacklisted" %
                               pkg.name)
                 return False
-            # a strict whitelist will not allow any changes not in the 
+            # a strict whitelist will not allow any changes not in the
             # whitelist, most people will want the relaxed whitelist
-            # that whitelists a package but pulls in the package 
+            # that whitelists a package but pulls in the package
             # dependencies
             strict_whitelist = apt_pkg.config.find_b(
                 "Unattended-Upgrade::Package-Whitelist-Strict", False)
             if (strict_whitelist and
-                not is_pkgname_in_whitelist(pkg.name, whitelist)):
+                    not is_pkgname_in_whitelist(pkg.name, whitelist)):
+
                 logging.debug("pkg '%s' package is not whitelisted" %
                               pkg.name)
                 return False
@@ -628,8 +649,8 @@ def conffile_prompt(destFile, prefix=""):
             obs = l[2]
         else:
             obs = None
-        # ignore if conffile is obsolete or does not exist
-        if obs == "obsolete" or not os.path.exists(prefix + conf_file):
+        # ignore if conffile is obsolete
+        if obs == "obsolete":
             continue
         # ignore state "newconffile" until its clearer if there
         # might be a dpkg prompt (LP: #936870)
@@ -641,10 +662,16 @@ def conffile_prompt(destFile, prefix=""):
             continue
         # record for later
         dpkg_status_conffiles[conf_file] = md5
-        # test against the installed file
-        with open(prefix + conf_file, 'rb') as fb:
-            current_md5 = apt_pkg.md5sum(fb)
+
+        # test against the installed file, if the local file got deleted
+        # by the admin thats ok but it may still trigger a conffile prompt
+        # (see debian #788049)
+        current_md5 = ""
+        if os.path.exists(prefix + conf_file):
+            with open(prefix + conf_file, 'rb') as fb:
+                current_md5 = apt_pkg.md5sum(fb)
         logging.debug("current md5: %s" % current_md5)
+
         # hashes are the same, no conffile prompt
         if current_md5 == md5:
             continue
@@ -723,7 +750,7 @@ def setup_apt_listchanges(conf="/etc/apt/listchanges.conf"):
         os.environ["APT_LISTCHANGES_FRONTEND"] = "none"
 
 
-def _send_mail_using_mailx(to_address, subject, body):
+def _send_mail_using_mailx(from_address, to_address, subject, body):
     # ensure that the body is a byte stream and that we do not
     # break on encoding errors (the default error mode is "strict")
     encoded_body = body.encode(
@@ -732,7 +759,7 @@ def _send_mail_using_mailx(to_address, subject, body):
     # unicode encoding errors (e.g. because the user is running a
     # ascii only system like the buildds)
     mail = subprocess.Popen(
-        [MAIL_BINARY, "-s", subject, to_address],
+        [MAIL_BINARY, "-r", from_address, "-s", subject, to_address],
         stdin=subprocess.PIPE, universal_newlines=False)
     mail.stdin.write(encoded_body)
     mail.stdin.close()
@@ -740,10 +767,11 @@ def _send_mail_using_mailx(to_address, subject, body):
     return ret
 
 
-def _send_mail_using_sendmail(to_address, subject, body):
+def _send_mail_using_sendmail(from_address, to_address, subject, body):
     # format as a proper mail
     msg = Message()
     msg['Subject'] = subject
+    msg['From'] = from_address
     msg['To'] = to_address
     # order is important here, Message() first, then Charset()
     #  then msg.set_charset()
@@ -772,18 +800,23 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, dpkg_log_content):
         return
     # if the operation was successful and the user has requested to get
     # mails on on errors, just exit here
-    if (res and apt_pkg.config.find_b(
+    if (res and
+        # see Debian Bug #703621
+        not re.search("^WARNING:", mem_log.getvalue(), re.MULTILINE) and
+        apt_pkg.config.find_b(
             "Unattended-Upgrade::MailOnlyOnError", False)):
         return
     # Check if reboot-required flag is present
+    reboot_flag_str = _(
+        "[reboot required]") if os.path.isfile(REBOOT_REQUIRED_FILE) else ""
+    # Check if packages are kept on hold
+    hold_flag_str = _("[package on hold]") if pkgs_kept_back else ""
     logging.debug("Sending mail to '%s'" % to_email)
-    if os.path.isfile(REBOOT_REQUIRED_FILE):
-        subject = _(
-            "[reboot required] unattended-upgrades result for '%s': %s") % (
-            host(), res)
-    else:
-        subject = _("unattended-upgrades result for '%s': '%s'") % (
-            host(), res)
+    subject = _(
+        "{hold_flag}{reboot_flag} unattended-upgrades result for "
+        "'{machine}': {result}").format(
+            hold_flag=hold_flag_str, reboot_flag=reboot_flag_str,
+            machine=host(), result=res).strip()
     body = _("Unattended upgrade returned: %s\n\n") % res
     if os.path.isfile(REBOOT_REQUIRED_FILE):
         body += _(
@@ -806,15 +839,17 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, dpkg_log_content):
     body += _("Unattended-upgrades log:\n")
     body += mem_log.getvalue()
 
+    from_email = apt_pkg.config.find("Unattended-Upgrade::Sender", "root")
+
     if os.path.exists(SENDMAIL_BINARY):
-        ret = _send_mail_using_sendmail(to_email, subject, body)
+        ret = _send_mail_using_sendmail(from_email, to_email, subject, body)
     elif os.path.exists(MAIL_BINARY):
-        ret = _send_mail_using_mailx(to_email, subject, body)
+        ret = _send_mail_using_mailx(from_email, to_email, subject, body)
     else:
         raise AssertionError(
             "This should never be reached, if we are here we either "
             "have sendmail or mailx. Please report this as a bug.")
-    logging.debug("mail returned: %s" % ret)
+    logging.debug("mail returned: %s", ret)
 
 
 def do_install(cache, pkgs_to_upgrade, blacklisted_pkgs, whitelisted_pkgs,
@@ -823,7 +858,7 @@ def do_install(cache, pkgs_to_upgrade, blacklisted_pkgs, whitelisted_pkgs,
     os.putenv("DEBIAN_FRONTEND", "noninteractive")
     setup_apt_listchanges()
 
-    logging.info(_("Writing dpkg log to '%s'") % logfile_dpkg)
+    logging.info(_("Writing dpkg log to '%s'"), logfile_dpkg)
 
     marked_delete = [pkg for pkg in cache.get_changes() if pkg.marked_delete]
     if marked_delete:
@@ -850,8 +885,7 @@ def do_install(cache, pkgs_to_upgrade, blacklisted_pkgs, whitelisted_pkgs,
                 options.verbose or options.debug)
         else:
             pkg_install_success = upgrade_normal(
-                cache, [pkg.name for pkg in pkgs_to_upgrade],
-                logfile_dpkg, options.verbose or options.debug)
+                cache, logfile_dpkg, options.verbose or options.debug)
     except Exception as e:
         # print unhandled exceptions here this way, while stderr is redirected
         os.write(2, ("Exception: %s\n" % e).encode('utf-8'))
@@ -934,18 +968,42 @@ def get_whitelisted_pkgs():
     return apt_pkg.config.value_list("Unattended-Upgrade::Package-Whitelist")
 
 
+def logged_in_users():
+    """Return a list of logged in users"""
+    # the "users" command always returns a single line with:
+    # "user1, user1, user2"
+    users = subprocess.check_output(
+        USERS, universal_newlines=True).rstrip('\n')
+    return set(users.split())
+
+
 def reboot_if_requested_and_needed(shutdown_lock=0):
-    # auto-reboot (if required and the config for this is set
-    if (apt_pkg.config.find_b(
-            "Unattended-Upgrade::Automatic-Reboot", False) and
-            os.path.exists(REBOOT_REQUIRED_FILE)):
-        when = apt_pkg.config.find(
-            "Unattended-Upgrade::Automatic-Reboot-Time", "now")
-        if shutdown_lock > 0:
-            os.close(shutdown_lock)
-        logging.warning("Found %s, rebooting" % REBOOT_REQUIRED_FILE)
-        # XXX: add time option
-        subprocess.call(["/sbin/shutdown", "-r", when])
+    """auto-reboot (if required and the config for this is set)"""
+    if not os.path.exists(REBOOT_REQUIRED_FILE):
+        return
+    if not apt_pkg.config.find_b(
+            "Unattended-Upgrade::Automatic-Reboot", False):
+        return
+    # see if we need to check for logged in users
+    if not apt_pkg.config.find_b(
+            "Unattended-Upgrade::Automatic-Reboot-WithUsers", True):
+        users = logged_in_users()
+        if users:
+            msg = gettext.ngettext(
+                "Found %s, but not rebooting because %s is logged in." % (
+                    REBOOT_REQUIRED_FILE, users),
+                "Found %s, but not rebooting because %s are logged in." % (
+                    REBOOT_REQUIRED_FILE, users),
+                len(users))
+            logging.warning(msg)
+            return
+    # reboot at the specified time
+    when = apt_pkg.config.find(
+        "Unattended-Upgrade::Automatic-Reboot-Time", "now")
+    if shutdown_lock > 0:
+        os.close(shutdown_lock)
+    logging.warning("Found %s, rebooting" % REBOOT_REQUIRED_FILE)
+    subprocess.call(["/sbin/shutdown", "-r", when])
 
 
 def write_stamp_file():
@@ -954,8 +1012,8 @@ def write_stamp_file():
         pass
 
 
-def try_to_upgrade(pkg, pkgs_to_upgrade, pkgs_kept_back, cache, allowed_origins,
-                     blacklisted_pkgs, whitelisted_pkgs):
+def try_to_upgrade(pkg, pkgs_to_upgrade, pkgs_kept_back, cache,
+                   allowed_origins, blacklisted_pkgs, whitelisted_pkgs):
 
     try:
         pkg.mark_upgrade()
@@ -979,7 +1037,7 @@ def try_to_upgrade(pkg, pkgs_to_upgrade, pkgs_kept_back, cache, allowed_origins,
         # can't upgrade
         logging.warning(
             _("package '%s' upgradable but fails to "
-                "be marked for upgrade (%s)") % (pkg.name, e))
+                "be marked for upgrade (%s)"), pkg.name, e)
         rewind_cache(cache, pkgs_to_upgrade)
         pkgs_kept_back.append(pkg.name)
 
@@ -995,9 +1053,10 @@ def calculate_upgradable_pkgs(cache, options, allowed_origins,
             logging.debug("Checking: %s (%s)" % (
                 pkg.name, getattr(pkg.candidate, "origins", [])))
         if (pkg.is_upgradable and
-              not is_pkgname_in_blacklist(pkg.name, blacklisted_pkgs) and
-              is_pkgname_in_whitelist(pkg.name, whitelisted_pkgs) and
-              is_allowed_origin(pkg.candidate, allowed_origins)):
+                not is_pkgname_in_blacklist(pkg.name, blacklisted_pkgs) and
+                is_pkgname_in_whitelist(pkg.name, whitelisted_pkgs) and
+                is_allowed_origin(pkg.candidate, allowed_origins)):
+
             try_to_upgrade(pkg,
                            pkgs_to_upgrade,
                            pkgs_kept_back,
@@ -1013,7 +1072,7 @@ def get_dpkg_log_content(logfile_dpkg, install_start_time):
         logfile_dpkg, install_start_time))
     content = []
     found_start = False
-    with open(logfile_dpkg) as fp:
+    with io.open(logfile_dpkg, encoding='utf-8') as fp:
         # read until we find the last "Log started: "
         for line in fp.readlines():
             # scan for the first entry we need (minimal-step mode
@@ -1028,6 +1087,44 @@ def get_dpkg_log_content(logfile_dpkg, install_start_time):
     return "".join(content)
 
 
+def get_auto_removable(cache):
+    return set([pkg.name for pkg in cache
+                if pkg.is_auto_removable])
+
+
+def do_auto_remove(cache, auto_removable, logfile_dpkg, verbose=False):
+    if not auto_removable:
+        return True
+
+    for pkgname in auto_removable:
+        logging.debug("marking %s for remove" % pkgname)
+        cache[pkgname].mark_delete()
+    logging.info(_("Packages that are auto removed: '%s'"),
+                 " ".join(sorted(auto_removable)))
+
+    # do it
+    res, error = cache_commit(cache, logfile_dpkg, verbose)
+    if res:
+        logging.info(_("Packages were successfully auto-removed"))
+    else:
+        logging.error(_("Auto-removing the packages failed!"))
+        logging.error(_("Error message: '%s'"), error)
+        logging.error(_("dpkg returned an error! See '%s' for details"),
+                      logfile_dpkg)
+    return res
+
+
+def clean_downloaded_packages(fetcher):
+    archivedir = os.path.dirname(
+        apt_pkg.config.find_dir("Dir::Cache::archives"))
+    for item in fetcher.items:
+        if os.path.dirname(os.path.abspath(item.destfile)) == archivedir:
+            try:
+                os.unlink(item.destfile)
+            except OSError:
+                pass
+
+
 def main(options, rootdir=""):
     # useful for testing
     if rootdir:
@@ -1052,7 +1149,7 @@ def main(options, rootdir=""):
     logging.info(_("Starting unattended upgrades script"))
 
     # display available origin
-    logging.info(_("Allowed origins are: %s") % allowed_origins)
+    logging.info(_("Allowed origins are: %s"), allowed_origins)
 
     # see debian #776752
     install_start_time = datetime.datetime.now().replace(microsecond=0)
@@ -1080,7 +1177,7 @@ def main(options, rootdir=""):
                     universal_newlines=True)
             except subprocess.CalledProcessError as e:
                 output = e.output
-            logging.warning(_("dpkg --configure -a output:\n%s" % output))
+            logging.warning(_("dpkg --configure -a output:\n%s"), output)
         else:
             logging.debug("Unclean dpkg state, but locked, another package "
                           "manager working?")
@@ -1110,7 +1207,7 @@ def main(options, rootdir=""):
         os.nice(-1)
         os.nice(20)
     except OSError as e:
-        if e.errno in (EPERM, EACCES):
+        if e.errno in (errno.EPERM, errno.EACCES):
             pass
         else:
             raise
@@ -1119,9 +1216,6 @@ def main(options, rootdir=""):
     actiongroup = apt_pkg.ActionGroup(cache._depcache)
     actiongroup  # pyflakes
 
-    # stuff in auto-remove state
-    pkgs_auto_removable = set([pkg.name for pkg in cache
-                               if pkg.is_auto_removable])
     # find out about the packages that are upgradable (in a allowed_origin)
     pkgs_to_upgrade, pkgs_kept_back = calculate_upgradable_pkgs(
         cache, options, allowed_origins, blacklisted_pkgs, whitelisted_pkgs)
@@ -1145,9 +1239,9 @@ def main(options, rootdir=""):
     try:
         pm.get_archives(fetcher, list, recs)
     except SystemError as e:
-        logging.error(_("GetArchives() failed: '%s'") % e)
+        logging.error(_("GetArchives() failed: '%s'"), e)
     res = fetcher.run()
-    logging.debug("fetch.run() result: %s" % res)
+    logging.debug("fetch.run() result: %s", res)
 
     if dpkg_conffile_prompt():
         # now check the downloaded debs for conffile conflicts and build
@@ -1156,18 +1250,18 @@ def main(options, rootdir=""):
             logging.debug("%s" % item)
             if item.status == item.STAT_ERROR:
                 print(_("An error occurred: '%s'") % item.error_text)
-                logging.error(_("An error occurred: '%s'") % item.error_text)
+                logging.error(_("An error occurred: '%s'"), item.error_text)
             if not item.complete:
                 print(_("The URI '%s' failed to download, aborting") %
                       item.desc_uri)
-                logging.error(_("The URI '%s' failed to download, aborting") %
+                logging.error(_("The URI '%s' failed to download, aborting"),
                               item.desc_uri)
                 sys.exit(1)
             if not os.path.exists(item.destfile):
-                print(_("Download finished, but file '%s' not there?!?" %
-                        item.destfile))
+                print(_("Download finished, but file '%s' not there?!?") %
+                      item.destfile)
                 logging.error("Download finished, but file '%s' not "
-                              "there?!?" % item.destfile)
+                              "there?!?", item.destfile)
                 sys.exit(1)
             if not item.is_trusted:
                 blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
@@ -1185,7 +1279,7 @@ def main(options, rootdir=""):
                           pkgname_from_deb(item.destfile))
                 # log to the logfile
                 logging.warning(_("Package '%s' has conffile prompt and "
-                                  "needs to be upgraded manually") %
+                                  "needs to be upgraded manually"),
                                 pkgname_from_deb(item.destfile))
                 blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
                 pkgs_kept_back.append(pkgname_from_deb(item.destfile))
@@ -1204,39 +1298,40 @@ def main(options, rootdir=""):
                 logging.debug("Checking the black and whitelist: %s" %
                               (pkg.name))
                 pkg.mark_upgrade()
-                if check_changes_for_sanity(cache, allowed_origins,
-                                            blacklisted_pkgs, whitelisted_pkgs):
+                if check_changes_for_sanity(cache,
+                                            allowed_origins,
+                                            blacklisted_pkgs,
+                                            whitelisted_pkgs):
                     pkgs_to_upgrade.append(pkg)
                 else:
                     if not (pkg.name in pkgs_kept_back):
                         pkgs_kept_back.append(pkg.name)
-                    logging.info(_("package '%s' not upgraded") % pkg.name)
+                    logging.info(_("package '%s' not upgraded"), pkg.name)
                     cache.clear()
                     for pkg2 in pkgs_to_upgrade:
                         pkg2.mark_upgrade()
     else:
         logging.debug("dpkg is configured not to cause conffile prompts")
 
-    # do auto-remove
+    # auto-removal
+    previous_autoremovals = get_auto_removable(cache)
     if apt_pkg.config.find_b(
             "Unattended-Upgrade::Remove-Unused-Dependencies", False):
-        now_auto_removable = set([pkg.name for pkg in cache
-                                  if pkg.is_auto_removable])
-        for pkgname in now_auto_removable - pkgs_auto_removable:
-            logging.debug("marking %s for remove" % pkgname)
-            cache[pkgname].mark_delete()
-        logging.info(_("Packages that are auto removed: '%s'") %
-                     " ".join(now_auto_removable - pkgs_auto_removable))
-
-    logging.debug("InstCount=%i DelCount=%i BrokenCount=%i" % (
-        cache._depcache.inst_count, cache._depcache.del_count,
-        cache._depcache.broken_count))
+        pending_autoremovals = previous_autoremovals
+    else:
+        pending_autoremovals = []
 
     # exit if there is nothing to do and nothing to report
-    if (len(pkgs_to_upgrade) == 0) and (len(pkgs_kept_back) == 0):
-        logging.info(_("No packages found that can be upgraded unattended"))
+    if (len(pending_autoremovals) == 0 and
+            len(pkgs_to_upgrade) == 0 and
+            len(pkgs_kept_back) == 0):
+        logging.info(_("No packages found that can be upgraded unattended "
+                       "and no pending auto-removals"))
         # FIXME: DRY violation, write_stamp_file() is used below as well
         write_stamp_file()
+        # check if we couldn't reboot on previous run because
+        # a user was logged-in at this time
+        reboot_if_requested_and_needed()
         return
 
     # check if its configured for install on shutdown, if so, the
@@ -1256,21 +1351,55 @@ def main(options, rootdir=""):
 
     # do the install based on the new list of pkgs
     pkgs = " ".join([pkg.name for pkg in pkgs_to_upgrade])
-    logging.info(_("Packages that will be upgraded: %s" % pkgs))
+    logging.info(_("Packages that will be upgraded: %s"), pkgs)
 
     # get log
     logfile_dpkg = os.path.join(_get_logdir(), 'unattended-upgrades-dpkg.log')
 
     # only perform install step if we actually have packages to install
-    pkg_install_success = None
+    pkg_install_success = True
     shutdown_lock = -1
     if len(pkgs_to_upgrade) > 0:
         # lock for the shutdown check - its fine if the system
         # is shutdown while downloading but not so much while installing
         shutdown_lock = apt_pkg.get_lock("/var/run/unattended-upgrades.lock")
         # do install
-        pkg_install_success = do_install(
-            cache, pkgs_to_upgrade, blacklisted_pkgs, whitelisted_pkgs, options, logfile_dpkg)
+        pkg_install_success = do_install(cache,
+                                         pkgs_to_upgrade,
+                                         blacklisted_pkgs,
+                                         whitelisted_pkgs,
+                                         options,
+                                         logfile_dpkg)
+
+    # now check if any auto-removing needs to be done
+    cache = UnattendedUpgradesCache(
+        rootdir=rootdir, allowed_origins=allowed_origins)
+    if cache._depcache.broken_count > 0:
+        print(_("Cache has broken packages, exiting"))
+        logging.error(_("Cache has broken packages, exiting"))
+        sys.exit(1)
+
+    # the user wants *all* auto-removals to be removed
+    if apt_pkg.config.find_b(
+            "Unattended-Upgrade::Remove-Unused-Dependencies", False):
+        auto_removals = get_auto_removable(cache)
+        pkg_install_success &= do_auto_remove(
+            cache, auto_removals, logfile_dpkg,
+            options.verbose or options.debug)
+    # the user wants *only new* auto-removals to be removed
+    elif apt_pkg.config.find_b(
+            "Unattended-Upgrade::Remove-New-Unused-Dependencies", True):
+        # calculate the new auto-removals
+        new_pending_autoremovals = get_auto_removable(cache)
+        auto_removals = new_pending_autoremovals - previous_autoremovals
+        pkg_install_success &= do_auto_remove(
+            cache, auto_removals, logfile_dpkg,
+            options.verbose or options.debug)
+
+    logging.debug("InstCount=%i DelCount=%i BrokenCount=%i"
+                  % (cache._depcache.inst_count,
+                     cache._depcache.del_count,
+                     cache._depcache.broken_count))
 
     # send a mail (if needed)
     if not options.dry_run:
@@ -1278,6 +1407,13 @@ def main(options, rootdir=""):
         send_summary_mail(
             pkgs, pkg_install_success, pkgs_kept_back, mem_log, log_content)
 
+    # clean after success install (if needed)
+    keep_key = "Unattended-Upgrade::Keep-Debs-After-Install"
+    if (not apt_pkg.config.find_b(keep_key, False) and
+            not options.dry_run and
+            pkg_install_success):
+        clean_downloaded_packages(fetcher)
+
     # FIXME: DRY violation, write_stamp_file() is used above as well
     # write timestamp file
     write_stamp_file()
